c       the x-files library for adaptive minimization

      subroutine qmodel (im,jm)
c---------------------------------------------------------------------
c pmodel - stores ALL solution model parameters in global arrays

c local variables:

c istot - total number of endmembers (excluding ordered species) used
c         to formulate the solution model in the input file
c jstot - total number of endmembers (excluding ordered species) used
c         in the computation (i.e., excluding missing endmembers), but
c         including dependent endmembers.
c kstot - total number of independet endmembers (excluding ordered species) 
c         used in the computation (i.e., excluding missing endmembers)
c insp(jstot/jstot+1) - pointer to original index of first kstot independent
c         endmembers, followed by jstot-kstot dependent endmembers, followed
c         by the ordered species, if present (jstot+1).
c kdsol(istot/istot+1) - endmember index, 0 if missing, -2 if dependent,
c          -1 if ordered species (istot+1/jstot+1), original indexing.
c jmsol(jstot,msp) - chemical species present on m'th site of jstot'th
c          endmember, original indexing.
c---------------------------------------------------------------------
      implicit none
  
      include 'perplex_parameters.h'

      integer im,istot,jstot,iend,isub,imd,mdep,jdep,
     *        ist,i,j,idep,ind,im2jm,
     *        iord,iterm,isite,isp,nsub1,nsub2,nttyp,
     *        insp,jmsol,nterm,nspm1,nsite,
     *        kstot,iddeps,jsmod,k,l,itic,kdsol,
     *        ii,jm,ict

      double precision wg,xmn,xmx,xnc,ecoef,acoef,smult,
     *                 a0,y2p,nu,r,depvnu,depenth,vlaar

c                                    bookkeeping variables
      integer ksmod, ksite, kmsol, knsp, mstot, inid

      common/ cxt0  /ksmod(l1),ksite(l1),kmsol(l1,m4,mst),knsp(l1,m4),
     *               mstot(l1),inid(l1,m4)
c                                    configurational entropy variables:
      integer msite, ksp, lterm, ksub1, ksub2, ktyp

      double precision qmult, d0, dcoef

      common/ cxt1i /msite(l1),ksp(l1,m10),lterm(l1,m10,m11),
     *               ksub1(l1,m10,m11,m0,m12),ksub2(l1,m10,m11,m0,m12),
     *               ktyp(l1,m10,m11,m0)
     *      / cxt1r /qmult(l1,m10),d0(l1,m10,m11),dcoef(l1,m10,m11,m0)
c                                    excess energy variables
      integer jterm, jord, jsub1, jsub2

      double precision wgl, scoef

      common/ cxt2i /jterm(l1),jord(l1),jsub1(l1,m1,m2),jsub2(l1,m1,m2)
     *      / cxt2r /wgl(l1,m1,m3),scoef(l1,m4)

      integer ideps

      double precision w,x,sdzdy,vlar,dvnu,deph,dydy

      common/ cxt3i /ideps(l1,2)
     *      / cxt3r /sdzdy(l1,m10,m11),
     *               vlar(l1,m4,m3),dvnu(l1,m4),deph(l1),dydy(l1,m4)
      
      integer ndep
      
      double precision y2pg

      common/ cxt4  /y2pg(l1,m4,m15),ndep(l1)
c                                    convert y -> x array
      integer indx

      common/ cxt5i /indx(l1,mst,msp,msp)
c                                    x coordinate description
      integer istg, ispg, imlt

      double precision xmng, xmxg, xncg

      common/ cxt6r /xmng(l1,mst,msp),xmxg(l1,mst,msp),xncg(l1,mst,msp)
      common/ cxt6i /istg(l1),ispg(l1,mst),imlt(l1,mst)

c                                    local input variables
      common/ cst108 /wg(m1,m3),xmn(mst,msp),xmx(mst,msp),xnc(mst,msp),
     *      iend(m4),isub(m1,m2,2),imd(mst),insp(m4),ist(mst),
     *      isp(mst),isite,iterm,iord,istot,jstot,kstot
     *      / cst107 /a0(m10,m11),acoef(m10,m11,m0),smult(m10),ecoef(m4)
     *      ,nsite,nspm1(m10),nterm(m10,m11),nsub1(m10,m11,m0,m12),
     *      nsub2(m10,m11,m0,m12),nttyp(m10,m11,m0)
     *      / cst221 /vlaar(m3,m4),jsmod
     *      / cst141 /depvnu(2),depenth,iddeps(2)
     *      / cst142 /jmsol(m4,mst),kdsol(m4)
     *      / cst146 /nu(m15,3),y2p(m4,m15),mdep,jdep(m15),idep(m15,3)
c                                make pointer from im (solution index) to
c                                storage index
      jm = jm + 1
c                                number of endmembers
      mstot(jm) = istot
c                                number of distinct identisites for entropy
      msite(jm) = nsite
c                                chemical mixing sites
      istg(jm) = isite
c                                number of species and multiplicity and
c                                site ranges
      do i = 1, isite
         ispg(jm,i) = isp(i)
         imlt(jm,i) = ist(i)
         do j = 1, isp(i) - 1
           xmng(jm,i,j) = xmn(i,j)
           xmxg(jm,i,j) = xmx(i,j)
           xncg(jm,i,j) = xnc(i,j)
         end do 
      end do 

c                                ----------------------------------------------
c                                save the excess terms.                              
      jterm(jm) = iterm
      jord(jm) = iord

      do i = 1, iterm 
         do j = 1, m3
            wgl(jm,i,j) = wg(i,j)
         end do 
c                                      high order terms not allowed for
c                                      van laar, why check here?
         if (iord.gt.2.and.jsmod.eq.3) call error (999,r,800,'PMODEL') 
         do j = 1, iord
c            jsub(jm,i,j) = isub(i,j,1)
c                                      replication for macroscopic format.
            jsub1(jm,i,j) = isub(i,j,1)
            jsub2(jm,i,j) = isub(i,j,2)
         end do 
      end do 
c                                ----------------------------------------------
c                                in endmembers and 
c                                endmember configurational entropies:
      do i = 1, istot
         scoef(jm,i) = ecoef(i)
         inid(jm,i) = kdsol(i) 
      end do
c                                ----------------------------------------------
c                                configurational entropy models

c                                site fractions as a function of bulk
c                                y's and dependent species y:

      do i = 1, nsite 
c                                     # of species, and site 
c                                     multiplicty.
         qmult(jm,i) = smult(i)
         ksp(jm,i) = nspm1(i) 
c                                     for each species, read
c                                     function to define the
c                                     site fraction of the species:
         do j = 1, nspm1(i)
c                                     # of terms in the 
c                                     site fraction function and a0.
            lterm(jm,i,j) = nterm(i,j)
            d0(jm,i,j) = a0(i,j)
c                                     for each term:
            do k = 1, nterm(i,j)
c                                     term type and coefficient:
               dcoef(jm,i,j,k) = acoef(i,j,k)
               ktyp(jm,i,j,k) = nttyp(i,j,k)

               do l = 1, nttyp(i,j,k)
c                                     note second index and nttyp>1 only
c                                     necessary for bragg-williams formulation

c                                     endmember indexes, used to be in hsub 
                  ksub1(jm,i,j,k,l) = nsub1(i,j,k,l) 
                  ksub2(jm,i,j,k,l) = nsub2(i,j,k,l) 
               end do 
            end do 
         end do
      end do 
c                                ----------------------------------------------
c                                van laar volumes, and pointers for "in" endmembers
      do i = 1, jstot
c                                if the solution is van laar save
c                                the "volume" function.
         if (jsmod.eq.3) then  
            do l = 1, m3
               vlar(jm,i,l) = vlaar(l,insp(i))
            end do 
         end if 
c                                insp points to the original position 
c                                of endmember i in the solution model
c                                input:
         knsp(jm,i) = insp(i)
      end do   
c                                create a y -> x array, this array is 
c                                to be used to convert endmember fractions (y's)
c                                back to geometric coordinates (x's).
      if (isite.gt.1) then 

         do i = 1, isite
            do j = 1, isp(i)
c                                now find all endmembers with
c                                species j on site i, this method
c                                is inefficient but idependent of
c                                endmember order.   
               ict = 0 
         
               do k = 1, istot
                  if (kdsol(k).ne.0.and.
     *                jmsol(k,i).eq.j) then
                     ict = ict + 1
                     indx(jm,i,j,ict) = k
                  end if 
               end do 
            end do 
         end do 
      end if 

      if (jsmod.eq.7.or.jsmod.eq.8) then 
c                                save y -> p array 
         ndep(jm) = mdep
         do i = 1, kstot 
            do j = 1, ndep(jm)
               y2pg(jm,i,j) = y2p(i,j)
            end do
         end do
      end if 
c                                ----------------------------------------------
c                                saved data with all endmembers regardless
c                                of whether they are present in the calculation:
      if (jsmod.eq.6.or.jsmod.eq.8) then 
c                                models with speciation:    
         deph(jm) = depenth

         do i = 1, 2
            ideps(jm,i) = iddeps(i)
         end do 
c                                first create derivatives of endmember
c                                fractions with respect to the ordered
c                                species:
         do i = 1, istot
            dydy(jm,i) = 0d0
         end do
 
         dydy(jm,istot+1) = 1d0

         do i = 1, 2
            dydy(jm,iddeps(i)) = -depvnu(i)
         end do

         if (jsmod.eq.8) then 
c                                pointer to ordered cpd
            knsp(jm,jstot+1) = insp(jstot+1)

c                                make an array to get the the fractions
c                                of the disordered species in the ordered
c                                species indexed by kstot counter rather
c                                than the istot counter
            do i = 1, kstot
               dvnu(jm,i) = 0d0

               do j = 1, 2
                  if (insp(i).ne.iddeps(j)) cycle
                  dvnu(jm,i) = depvnu(j)
               end do 
            end do  

         end if
c                                site fractions as a function of bulk
c                                y's and dependent species y:
         do i = 1, nsite 
c                                     for each species, read
c                                     function to define the
c                                     site fraction of the species:
            do j = 1, nspm1(i)
c                                     # of terms in the 
c                                     site fraction function and a0.
               lterm(jm,i,j) = nterm(i,j)
               d0(jm,i,j) = a0(i,j)
               sdzdy(jm,i,j) = 0d0
c                                     for each term:
               do k = 1, nterm(i,j)
c                                     term type and coefficient:
                  dcoef(jm,i,j,k) = acoef(i,j,k)

                  itic = 0 

c                                     endmember indexes  
                  ind = nsub1(i,j,k,1) 
c                 hsub(jm,i,j,k) = ind
c                                     get derivatives, this
c                                     may not always work:
                  do ii = 1, 2
                     if (ind.eq.iddeps(ii)) then
                        sdzdy(jm,i,j) = sdzdy(jm,i,j) 
     *                                + dydy(jm,iddeps(ii))*acoef(i,j,k)  
                        itic = itic + 1
c                                      high order terms not allowed
                        if (itic.gt.1) call error (999,r,801,'PMODEL') 
                     end if                       
                  end do

                  if (ind.eq.istot+1) 
     *               sdzdy(jm,i,j) = sdzdy(jm,i,j) + acoef(i,j,k)  
                
               end do 
            end do 
         end do
      else 
c
 
      end if 

999   end 

      subroutine gsolut (ids,id,z)
c-----------------------------------------------------------------------
c gsolut computes the total (excess+ideal) free energy of solution 
c for a solution identified by index ids and composition z(i,j),
c where i indexes the site, j the species. 

c here it is assumed that the compositions have been filtered to 
c eliminate invalid compositions. 

c ksmod(ids)* - indicates model type (= jsmod flag)
c mstot(ids) - the total number of endmembers in the model
c ksite(ids)* - number of mixing sites
c ksp(ids,ksite)* - number of species on each site
c kmsol(ids,istot,ksite)* - ksite indices for site populations of 
c                          endmember isot 
c inid(ids,istot) - endmember id, if zero missing.
c-----------------------------------------------------------------------
      implicit none
 
      include 'perplex_parameters.h'

      double precision z(mst,msp), zpt(m4)
c                                    excess energy variables
      integer jterm, jord, jsub1, jsub2

      double precision wgl, scoef

      common/ cxt2i /jterm(l1),jord(l1),jsub1(l1,m1,m2),jsub2(l1,m1,m2)
     *      / cxt2r /wgl(l1,m1,m3),scoef(l1,m4)
     *      / cxt3i /ideps(l1,2)
     *      / cxt3r /sdzdy(l1,m10,m11),
     *               vlar(l1,m4,m3),dvnu(l1,m4),deph(l1),dydy(l1,m4)
      
      character fname 

      integer ipoint,j,imyn,jm,m,
     *        ideps,ifp,jend,k,ixp,jd,
     *        hsub,kterm,kord,lsub,id,ndep,jsub,ixpt,l,
     *        ids,ikp,ivarrx,ivarip,isudo,ivar

      double precision r,vlar,deph,dydy,y2pg,gph,
     *                 sdzdy,x,dvnu,tr,pr,ps,sxs,gzero,p0ord,
     *                 dg,exces,p,t,xco2,u1,u2

      common/ cst304 /sxs(k13),exces(m3,k1),jend(h9,k12),ixp(k1),ifp(k1)
      common/ cst60  /ipoint,imyn
     *      / cst61  /ikp(k1),ivarrx(k2),ivarip(k2),isudo,ivar
     *      / cst5   /p,t,xco2,u1,u2,tr,pr,r,ps/ csta7 /fname(h9)
c------------------------------------------------------------------------
c                                    bookkeeping variables
      integer ksmod, ksite, kmsol, knsp, mstot, inid

      common/ cxt0  /ksmod(l1),ksite(l1),kmsol(l1,m4,mst),knsp(l1,m4),
     *               mstot(l1),inid(l1,m4)

c                                    configurational entropy variables:
      integer msite, ksp, lterm, ksub1, ksub2, ktyp

      double precision qmult, d0, dcoef

      common/ cxt1i /msite(l1),ksp(l1,m10),lterm(l1,m10,m11),
     *               ksub1(l1,m10,m11,m0,m12),ksub2(l1,m10,m11,m0,m12),
     *               ktyp(l1,m10,m11,m0)
     *      / cxt1r /qmult(l1,m10),d0(l1,m10,m11),dcoef(l1,m10,m11,m0)

     *      / cxt4  /y2pg(l1,m4,m15),ndep(l1)
c                                     working arrays for solutions:
      double precision pa, p0a,w, y, x

      common / cst5r /p0a(m4),pa(m4),y(m4),w(m1),x(mst,msp)
c------------------------------------------------------------------------
c                              --------------------------------------
c                              composition check, here we check that
c                              the composition is in bounds and does
c                              not involve a missing endmember
      do l = 1, mstot(ids)
c                              pa is the holland and powell porportion
         pa(l) = 0d0
c                              zpt is the bragg-williams fraction
         zpt(l) = 1d0
         do m = 1, ksite(ids)
c                              check for invalid compositions, this
c                              is almost certainly not necessary
            call zchk (z(m,kmsol(ids,l,m)),fname(ids))
            zpt(l) = zpt(l)*z(m,kmsol(ids,l,m))
         end do
c                              zp is the mole fraction of endmember l
         if (zpt(l).gt.0.9999d0.and.inid(ids,l).gt.0) goto 999
c                              reject the compositions if missing
c                              endmember
         if (inid(ids,l).eq.0.and.zpt(l).ne.0d0) goto 999
      end do  
c                              ---------------------------------------
c                              composition is ok, next some bookkeeping
c      kphct = kphct + 1
c      kkp(kphct) = ids

      if (ksmod(id).eq.1) then 
c                                 microscopic formulation
         call smix0 (z,dg,id)

      else if (ifp(id).eq.6) then
c                                 speciation model with 1 order parameter
         ids = ikp(id)

         jd = mstot(jm) + 1

         do j = 1, mstot(jm)
            y(j) = 0d0
         end do 
c                                 initialize gph and get any dqf corrections
         call gexces (id,gph)
c                                 get gmech
         do k = 1, jend(ikp(id),1) 
            call gcpd (jend(ikp(id),2+k),dg)
            y(knsp(jm,k)) = sxs(ixp(id)+k)
            gph = gph + dg*sxs(ixp(id)+k)
         end do 
c                                 compute margules coefficients
         do j = 1, jterm(jm)
            w(j) = wgl(jm,j,1) + t * wgl(jm,j,2) + p * wgl(jm,j,3)
         end do 

c                                 get the speciation energy effect
         call specis (dg,jm,jd)
c                                 subtract endmember sconfg
         gph = gph + dg + t * exces(2,id)

      else if (ifp(id).eq.8) then 
c                                 reciprocal solution speciation model 
c                                 with 1 order parameter
         ids = ikp(id)

         jd = mstot(jm) + 1
c                                 initialize x's and p0's
         do j = 1, mstot(jm)
            y(j) = 0d0
            p0a(j) = 0d0
         end do 
c                                 load x's from sxs array.                              
         do j = 1, jend(ids,1) 
            y(knsp(jm,j)) = sxs(ixp(id)+j)
         end do 
c                                 convert y's to p's, this loop 
c                                 could be eliminated by saving the
c                                 p's.
         ixpt = ixp(id) + jend(ids,1) - ndep(jm)
         p0ord = 0d0

         do k = 1, jend(ids,1) - ndep(jm)

            p0a(knsp(jm,k)) =  sxs(ixp(id)+k)

            do l = 1, ndep(jm)
               p0a(knsp(jm,k)) =  p0a(knsp(jm,k)) 
     *                           + y2pg(jm,k,l) * sxs(ixpt+l)
            end do 

            p0ord = p0ord + p0a(knsp(jm,k))

         end do 
c                                 the initial fraction of the ordered species
         p0ord = 1d0 - p0ord
         p0a(knsp(jm,jend(ids,1)+1)) = p0ord
c                                 compute margules coefficients
         do j = 1, jterm(jm)
            w(j) = wgl(jm,j,1) + t * wgl(jm,j,2) + p * wgl(jm,j,3)
         end do 
c                                 get the speciation energy effect
         call specip (gph,jm,jd)
c                                 get endmember sconfg and dqf's
         call gexces (id,dg) 

         gph = gph + dg
c                                 get gmech
         do k = 1, jend(ids,1) - ndep(jm)
            call gcpd (jend(ids,2+k),dg)
            gph = gph + dg * (p0a(knsp(jm,k)) + dvnu(jm,k)*p0ord)
         end do 

      else 
c                              a pseudocompound without speciation:
         if (ifp(id).ne.1) then
c                              get the excess and/or ideal mixing effect
c                              and/or dqf corrections:
            if (ifp(id).eq.23) then 
               call toop(id,gph)
            else if (ifp(id).eq.26) then 
               call hcneos (gph,sxs(ixp(id)+1),
     *                      sxs(ixp(id)+2),sxs(ixp(id)+3))
            else
               call gexces (id,gph)
            end if 
c                              compute mech mix G for 
c                              all models except fluid 
            do k = 1, jend(ikp(id),1) 
               call gcpd (jend(ikp(id),2+k),dg)
               gph = gph + dg*sxs(ixp(id)+k)
            end do 

         else 
c                              get the excess and/or ideal mixing effect
c                              and/or dqf corrections:
            call fexces (id,gph)
c                              excess props don't include vdp:
            do k = 1, jend(ikp(id),1) 
               gph = gph + gzero(jend(ikp(id),2+k))*sxs(ixp(id)+k)
            end do 
         end if 
c                              for van laar get fancier excess function         
         if (ifp(id).eq.3) call gvlaar (ikp(id),id,gph)
         
      end if 

999   end

      subroutine smix0 (x,dlnw,id)
c----------------------------------------------------------------------
c subroutine to evaluate the log of the number of site configurations
c per mole (dlnw) of a solution id with composition x. bragg-williams 
c version. Oct '04.
c----------------------------------------------------------------------
      implicit none

      include 'perplex_parameters.h'

      integer ksmod, ksite, kmsol, knsp, mstot, inid

      common/ cxt0  /ksmod(l1),ksite(l1),kmsol(l1,m4,mst),knsp(l1,m4),
     *               mstot(l1),inid(l1,m4)

      double precision x(mst,msp), z, zt, dlnw, zad

      integer i,j,msite,lterm,ksp,k,id,l

      double precision qmult, d0, dcoef         

      integer ksub1, ksub2, ktyp

      common/ cxt1i /msite(l1),ksp(l1,m10),lterm(l1,m10,m11),
     *               ksub1(l1,m10,m11,m0,m12),ksub2(l1,m10,m11,m0,m12),
     *               ktyp(l1,m10,m11,m0)
     *      / cxt1r /qmult(l1,m10),d0(l1,m10,m11),dcoef(l1,m10,m11,m0)

      dlnw = 0d0
c                                     for each site
      do i = 1, ksite(id)

         zt = 0d0
c                                     get site fractions
         do j = 1, ksp(id,i)
            z = d0(id,i,j) 
c                                     for each term:
            do k = 1, lterm(id,i,j)
               zad = dcoef(id,i,j,k)
               do l = 1, ktyp(id,i,j,k)
                  zad = zad * x(ksub1(id,i,j,k,l),ksub2(id,i,j,k,l))
               end do 
               z = z + zad
            end do  

            zt = zt + z
            if (z.gt.0d0) dlnw = dlnw - qmult(id,i) * z * dlog(z)
            
         end do 

         z = 1d0 - zt
         if (z.gt.0d0) dlnw = dlnw - qmult(id,i) * z * dlog(z)

      end do 

      end

      function nsconfg (id)
c----------------------------------------------------------------------
c subroutine to evaluate configurational entropy of a solution
c with composition y. 
c----------------------------------------------------------------------

      implicit none 

      include 'perplex_parameters.h'

      double precision qmult, d0, dcoef, r, tr, pr, ps, zt, dlnw,
     *                 nsconfg, z, v   

      integer ksub1, ksub2, ktyp, ksp, i, j, k, msite, id, lterm      

      common/ cxt1i /msite(l1),ksp(l1,m10),lterm(l1,m10,m11),
     *               ksub1(l1,m10,m11,m0,m12),ksub2(l1,m10,m11,m0,m12),
     *               ktyp(l1,m10,m11,m0)
     *      / cxt1r /qmult(l1,m10),d0(l1,m10,m11),dcoef(l1,m10,m11,m0)
     *      / cst5   /v(l2),tr,pr,r,ps
c                                     working arrays for solutions:
      double precision pa, p0a,w, y

      common / cst5x /p0a(m4),pa(m4),y(m4),w(m1)

      dlnw = 0d0
c                                     for each site
      do i = 1, msite(id)

         zt = 0d0
c                                     get site fractions
         do j = 1, ksp(id,i)
            z = d0(id,i,j)
c                                     for each term:
            do k = 1, lterm(id,i,j)
               z = z + dcoef(id,i,j,k) * y(ksub1(id,i,j,k,1))
            end do 

            zt = zt + z
            if (z.gt.0d0) dlnw = dlnw - qmult(id,i) * z * dlog (z)

         end do 
c                                     the last species is by difference:
         z = 1d0 - zt 
         if (z.gt.0d0) dlnw = dlnw - qmult(id,i) * z * dlog(z)

      end do 

      nsconfg = dlnw*r

      end

      subroutine ytox (ids)
c----------------------------------------------------------------------
c subroutine to convert endmember fractions (y) to geometric coorinates 
c (x) for solution model ids.
c----------------------------------------------------------------------
      implicit none 

      include 'perplex_parameters.h'

      integer ids
c                                    convert y -> x array
      integer indx

      common/ cxt5i /indx(l1,mst,msp,msp)
c                                    working arrays for solutions:
      double precision pa, p0a,w, y, x

      common / cst5r /p0a(m4),pa(m4),y(m4),w(m1),x(mst,msp)
c                                    x coordinate description
      integer istg, ispg, imlt

      common/ cxt6i /istg(l1),ispg(l1,mst),imlt(l1,mst)

      do i = 1, istg(ids)
         do j = 1, ispg(ids,i)
            x(i,j) = 0d0
         end do 
         do j = 1, ispg(ids,i)
            x(i,j) = x(i,j) + y(indx(ids,i,
         end do 


            